@@ -254,6 +254,11 @@ static const int protocol_priority[] = {
   0
 };
 
+static const int dtls_protocol_priority[] = {
+  GNUTLS_DTLS1_0,
+  0
+};
+
 static const int protocol_priority_suiteb[] = {
   GNUTLS_TLS1_2,
   0
@@ -289,9 +294,13 @@ static const int kx_priority_secure[] = {
    */
   GNUTLS_KX_ECDHE_ECDSA,
   GNUTLS_KX_ECDHE_RSA,
+  /* KX-RSA is now ahead of DHE-RSA and DHE-DSS due to the compatibility
+   * issues the DHE ciphersuites have. That is, one cannot enforce a specific
+   * security level without dropping the connection. 
+   */
+  GNUTLS_KX_RSA,
   GNUTLS_KX_DHE_RSA,
   GNUTLS_KX_DHE_DSS,
-  GNUTLS_KX_RSA,
   /* GNUTLS_KX_ANON_DH: Man-in-the-middle prone, don't add!
    * GNUTLS_KX_RSA_EXPORT: Deprecated, don't add!
    */
@@ -808,6 +817,11 @@ gnutls_priority_init (gnutls_priority_t * priority_cache,
                   bulk_fn (&(*priority_cache)->protocol,
                                  protocol_priority);
                 }
+              else if (strncasecmp (&broken_list[i][1], "VERS-DTLS-ALL", 13) == 0)
+                {
+                  bulk_fn (&(*priority_cache)->protocol,
+                                 dtls_protocol_priority);
+                }
               else
                 {
                   if ((algo =
@@ -911,6 +925,7 @@ gnutls_priority_init (gnutls_priority_t * priority_cache,
             {
               (*priority_cache)->no_padding = 1;
               (*priority_cache)->allow_large_records = 1;
+              (*priority_cache)->allow_wrong_pms = 1;
             }
           else if (strcasecmp (&broken_list[i][1], "NO_EXTENSIONS") == 0)
             {
